module	DJ #(
	parameter FFT_IWIDTH = 16,
	parameter FFT_OWIDTH = 21,
	parameter IFFT_IWIDTH = 21,
	parameter IFFT_OWIDTH = 26,
	parameter LGWIDTH = 9
)(
	input clk,  //  Master Clock
	input reset,

	// Effects Controller Avalon Slave
	input		chipselect,
	input [7:0]	effects_writedata,
	input		effects_write,
	output [7:0]	effects_readdata,
	input		effects_read,
	input [5:0]	effects_address,


	// Left Channel Input Buffer 
	input signed [FFT_IWIDTH - 1:0]	left_fifo_source_stream, // to fft block
	input				left_fifo_source_valid,
	output				left_fifo_source_ready,


	// Right Channel Input Buffer
	input signed [FFT_IWIDTH - 1:0]	right_fifo_source_stream, // to fft block
	input				right_fifo_source_valid,
	output				right_fifo_source_ready,

	// 

	// Left Channel Output Codec Buffer
	output signed [FFT_IWIDTH - 1:0]	codec_left_fifo_sink_stream, // from fft bus
	output					codec_left_fifo_sink_valid,
	input					codec_left_fifo_sink_ready,


	// Right Channel Output Codec Buffer
	output signed [FFT_IWIDTH - 1:0]	codec_right_fifo_sink_stream, // from fft bus
	output					codec_right_fifo_sink_valid,
	input					codec_right_fifo_sink_ready
);
	/****************** 
		LEFT DECLARATIONS
	********************/

	// FFT Block 
	wire				left_fft_ce;
	wire [(2 * FFT_OWIDTH - 1):0]	left_fft_o_result;
	wire				left_fft_o_sync;

	// IFFT Block
	wire				left_ifft_i_ce;
	wire [(2 * IFFT_IWIDTH - 1):0]	left_ifft_i_sample;
	reg				left_ifft_i_reset;
	wire [2 * IFFT_OWIDTH - 1:0]	left_ifft_o_result;
	wire				left_ifft_o_sync;

	// LPF Block
	reg [(LGWIDTH - 1):0]		left_lpf_cutoff;
	wire				left_lpf_o_ce, left_lpf_o_sync;
	wire [(2 * IFFT_IWIDTH - 1):0]	left_lpf_o_sample;

	// HPF Block
	reg [(LGWIDTH - 1):0]		left_hpf_cutoff;
	wire				left_hpf_o_ce, left_hpf_o_sync;
	wire [(2 * IFFT_IWIDTH - 1):0]	left_hpf_o_sample;


	/****************** 
		RIGHT DECLARATIONS
	********************/
	/*
	// FFT Block 
	wire				right_fft_ce;
	wire [(2 * FFT_OWIDTH - 1):0]	right_fft_o_result;
	wire				right_fft_o_sync;

	// IFFT Block
	wire				right_ifft_i_ce;
	wire [(2 * IFFT_IWIDTH - 1):0]	right_ifft_i_sample;
	reg				right_ifft_i_reset;
	wire [2 * IFFT_OWIDTH - 1:0]	right_ifft_o_result;
	wire				right_ifft_o_sync;

	// LPF Block
	reg [(LGWIDTH):0]		right_lpf_cutoff;
	wire				right_lpf_o_ce, right_lpf_o_sync;
	wire [(2 * IFFT_IWIDTH - 1):0]	right_lpf_o_sample;
	*/

	logic do_left_filtering;
	// logic do_right_filtering;

	/****************** 
		MIXER LOGIC
	********************/

	logic async_clr;

	//fifo wires for data from mixer module (unfiltered)
	logic left_song_read_req;
	logic right_song_read_req;

	logic left_song_almost_empty;
	logic right_song_almost_empty;

	logic left_song_almost_full;
	logic right_song_almost_full;

	logic left_song_empty;
	logic right_song_empty;

	logic left_song_full;
	logic right_song_full;

	logic [10:0] left_song_usedw;
	logic [10:0] right_song_usedw;

	logic [15:0] left_song_fifo_stream;
	logic [15:0] right_song_fifo_stream;


	fifo_interrupt left_song_data(.aclr(async_clr), .clock(clk), .data(left_fifo_source_stream),
		.rdreq(left_song_read_req), .wrreq(left_fifo_source_valid && right_fifo_source_valid),
		.almost_empty(left_song_almost_empty), .almost_full(left_song_almost_full),
		.empty(left_song_empty), .full(left_song_full), .q(left_song_fifo_stream), .usedw(left_song_usedw)); 

	fifo_interrupt right_song_data(.aclr(async_clr), .clock(clk), .data(right_fifo_source_stream),
		.rdreq(right_song_read_req), .wrreq(left_fifo_source_valid && right_fifo_source_valid),
		.almost_empty(right_song_almost_empty), .almost_full(right_song_almost_full),
		.empty(right_song_empty), .full(right_song_full), .q(right_song_fifo_stream), .usedw(right_song_usedw));

	reg [10:0] counter;

	//write data from fifos to codec
	always @(posedge clk) begin
		if (reset) begin
			async_clr <= 1;
		end
		if (codec_left_fifo_sink_ready && codec_right_fifo_sink_ready && counter < 1020) begin
			codec_left_fifo_sink_valid <= 0;
			codec_right_fifo_sink_valid <= 0;
			left_result_read_req <= 0;
			right_result_read_req <= 0;
			counter <= counter + 1;
			async_clr <= 0;
		end else if (!left_result_empty && !right_result_empty && codec_left_fifo_sink_ready && codec_right_fifo_sink_ready && counter == 1020) begin
			codec_left_fifo_sink_valid <= 1;
			codec_right_fifo_sink_valid <= 1;
			left_result_read_req <= 1;
			right_result_read_req <= 1;
			counter <= 0;
			async_clr <= 0;
		end else begin
			codec_left_fifo_sink_valid <= 0;
			codec_right_fifo_sink_valid <= 0;
			left_result_read_req <= 0;
			right_result_read_req <= 0;
			async_clr <= 0;
		end
	end


	/****************** 
		LEFT LOGIC
	********************/


	// only allow pipeline to proceed if values are available at the input 
	// AND the output buffer is not at risk of overflowing
	assign left_fft_ce = do_left_filtering && !left_song_empty && !left_result_almost_full;
	assign left_song_read_req = !left_song_empty && !left_result_almost_full;
	//allow mixer module to write data to song fifo
	assign left_fifo_source_ready = !left_song_almost_full;

	//codec left DAC output stream
	assign codec_left_fifo_sink_stream = left_result_stream + right_result_stream; 

	wire	left_w_syncd;
	reg	left_r_syncd;

	// Synchronize and latch FFT windows with IFFT
	always @(posedge clk) begin
		if (reset)
			left_r_syncd <= 1'b0;
		else
			left_r_syncd <= left_r_syncd || left_fft_o_sync;
	end

	assign	left_w_syncd = left_r_syncd || left_fft_o_sync;

	// Generate FFT Block for left song
	fftmain left_fft(
		.i_clk(clk),
		.i_reset(reset),
		.i_ce(left_fft_ce),
		.i_sample({left_song_fifo_stream, 16'b0}),
		.o_result(left_fft_o_result),
		.o_sync(left_fft_o_sync));

	// Generate LPF Block
	filter #(
		.LGWIDTH(LGWIDTH),
		.LGFILTWIDTH(5),
		.COEFFILE("lpf.hex")
	) left_lpf (   
		.i_cutoff(left_lpf_cutoff),
		.i_clk(clk),
		.i_ce(((left_fft_ce)&&(left_w_syncd))),
		.i_sample(left_fft_o_result), 
		.i_sync(left_fft_o_sync),
		.o_ce(left_lpf_o_ce),
		.o_sample(left_lpf_o_sample), 
		.o_sync(left_lpf_o_sync));

	filter #(
		.LGWIDTH(LGWIDTH),
		.LGFILTWIDTH(5),
		.COEFFILE("hpf.hex")
	) left_hpf (   
		.i_cutoff(left_hpf_cutoff),
		.i_clk(clk),
		.i_ce(left_lpf_o_ce),
		.i_sample(left_lpf_o_sample), 
		.i_sync(left_lpf_o_sync),
		.o_ce(left_hpf_o_ce),
		.o_sample(left_hpf_o_sample), 
		.o_sync(left_hpf_o_sync));

	assign left_ifft_i_ce = left_hpf_o_ce;
	assign left_ifft_i_sample = left_hpf_o_sample;

	// Generate IFFT Block
	ifftmain left_ifft(
		.i_clk(clk),
		.i_reset(reset),
		.i_ce(left_ifft_i_ce),
		.i_sample(left_ifft_i_sample),
		.o_result(left_ifft_o_result),
		.o_sync(left_ifft_o_sync));

	logic left_samples_ready;

	//check when first sample is ready from the ifft
	always @(posedge clk) begin
		if (reset) begin
			left_samples_ready <= 0;
		end else if(left_ifft_o_sync) begin
			left_samples_ready <= 1;
		end
	end

	//wires for fifos that store left ifft results
	logic left_result_read_req;
	logic left_result_almost_empty;
	logic left_result_almost_full;
	logic left_result_empty;
	logic left_result_full;
	logic [15:0] left_result_stream;
	logic [10:0] left_usedw;

	logic left_result_ready;
	assign left_result_ready = left_samples_ready && left_ifft_i_ce;

	fifo_interrupt left_result(.aclr(async_clr), .clock(clk),
		.data(do_left_filtering ? left_ifft_o_result[FFT_IWIDTH + IFFT_OWIDTH - 1: IFFT_OWIDTH] : left_song_fifo_stream),
		.rdreq(left_result_read_req), .wrreq(do_left_filtering ? left_result_ready : left_song_read_req),
		.almost_empty(left_result_almost_empty), .almost_full(left_result_almost_full),
		.empty(left_result_empty), .full(left_result_full), .q(left_result_stream), .usedw(left_usedw));



	/****************** 
		RIGHT LOGIC
	********************/


	// only allow pipeline to proceed if values are available at the input 
	// AND the output buffer is not at risk of overflowing
	// assign right_fft_ce = do_right_filtering && !right_song_empty && !right_result_almost_full;
	assign right_song_read_req = !right_song_empty && !right_result_almost_full;

	assign right_fifo_source_ready = !right_song_almost_full;

	assign codec_right_fifo_sink_stream =  left_result_stream  + right_result_stream;

	/*
	wire	right_w_syncd;
	reg	right_r_syncd;

	// Synchronize and latch FFT windows with IFFT
	always @(posedge clk) begin
		if (reset)
			right_r_syncd <= 1'b0;
		else
			right_r_syncd <= right_r_syncd || right_fft_o_sync;
	end

	assign	right_w_syncd = right_r_syncd || right_fft_o_sync;
	 

	// Generate FFT Block for right song
	fftmain right_fft(
		.i_clk(clk),
		.i_reset(reset),
		.i_ce(right_fft_ce),
		.i_sample({right_song_fifo_stream, 16'b0}),
		.o_result(right_fft_o_result),
		.o_sync(right_fft_o_sync));

	assign right_lpf_cutoff = 10'd40; 

	// Generate LPF Block
	filter #(
		.LGWIDTH(LGWIDTH),
		.LGFILTWIDTH(5),
		.COEFFILE("lpf.hex")
	) left_lpf (   
		.i_cutoff(right_lpf_cutoff),
		.i_clk(clk),
		.i_ce(((right_fft_ce)&&(right_w_syncd))),
		.i_sample(right_fft_o_result), 
		.i_sync(right_fft_o_sync),
		.o_ce(right_lpf_o_ce),
		.o_sample(right_lpf_o_sample), 
		.o_sync(right_lpf_o_sync));

	assign right_hpf_cutoff = 10'd10; 

	filter #(
		.LGWIDTH(LGWIDTH),
		.LGFILTWIDTH(5),
		.COEFFILE("hpf.hex")
	) right_hpf (   
		.i_cutoff(right_hpf_cutoff),
		.i_clk(clk),
		.i_ce(right_lpf_o_ce),
		.i_sample(right_lpf_o_sample), 
		.i_sync(right_lpf_o_sync),
		.o_ce(right_hpf_o_ce),
		.o_sample(right_hpf_o_sample), 
		.o_sync(right_hpf_o_sync));

	assign right_ifft_i_ce = right_hpf_o_ce;
	assign right_ifft_i_sample = right_hpf_o_sample;

	// Generate IFFT Block
	ifftmain right_ifft(
		.i_clk(clk),
		.i_reset(reset),
		.i_ce(right_ifft_i_ce),
		.i_sample(right_ifft_i_sample),
		.o_result(right_ifft_o_result),
		.o_sync(right_ifft_o_sync));
	    
	logic right_samples_ready;

	always @(posedge clk) begin
		if (reset) begin
			right_samples_ready <= 0;
		end else if(right_ifft_o_sync) begin
			right_samples_ready <= 1;
		end
	end
	*/

	logic right_result_read_req;
	logic right_result_almost_empty;
	logic right_result_almost_full;
	logic right_result_empty;
	logic right_result_full;
	logic [15:0] right_result_stream;
	logic [10:0] right_usedw;

	logic right_result_ready;
	// assign right_result_ready = right_samples_ready && right_ifft_i_ce;

	fifo_interrupt right_result(.aclr(async_clr), .clock(clk), .data(right_song_fifo_stream),
		.rdreq(right_result_read_req), .wrreq(right_song_read_req),
		.almost_empty(right_result_almost_empty), .almost_full(right_result_almost_full),
		.empty(right_result_empty), .full(right_result_full), .q(right_result_stream), .usedw(right_usedw));

	/*reg [2:0] ctr;

	wire clk_div;
	assign clk_div = ctr[2];

	always @(posedge clk) ctr <= ctr + 1;*/


	// assign codec_left_fifo_sink_ready = (ctr == 0) ? 1'b1 : 1'b0;

	// assign codec_right_fifo_source_ready = codec_left_fifo_source_ready;

	/*********************
		EFFECTS
	*********************/

	always @(posedge clk) begin
		if (reset) begin
			left_lpf_cutoff <= 9'd40;
			left_hpf_cutoff <= 9'd10;
			do_left_filtering <= 0;
		end else if (chipselect && effects_write) begin
			case (effects_address) 
			6'd0: do_left_filtering <= effects_writedata;
			6'd1: left_lpf_cutoff <= effects_writedata;
			6'd2: left_hpf_cutoff <= effects_writedata;
			default: ;
			endcase
		end
	end  
endmodule
