/*
 * Userspace program that communicates with the vga_ball device driver
 * through ioctls
 *
 * Stephen A. Edwards
 * Columbia University
 */

#include <assert.h>
#include <endian.h>
#include <fcntl.h>
#include <stdint.h>
#include <stdio.h>
#include "mixer.h"
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <unistd.h>
#include "usbkeyboard.h"
#include <pthread.h>
#include <stdlib.h>
#include "usbhidkeys.h"
#include <sys/stat.h>

#define GAIN_UP 1.1
#define GAIN_DOWN 0.9
#define LE16TOH(x) (le16toh(*(uint16_t *)(x)))
#define LE32TOH(x) (le32toh(*(uint32_t *)(x)))
#define FMT_OF(wi) (wi.file + wi.fmt_offset)
#define DATA_OF(wi) (wi.file + wi.data_offset)
#define SAMPLE_AT(wi, i) \
	(*(int16_t *)(DATA_OF(wi) + 8 + wi.bytes_per_whole_sample*(i)))
#define SAMPLE_AT_WO(wi, i) SAMPLE_AT(wi, (i + wi.song_offset)%wi.num_samples)
#define MIN(a, b) ((a) < (b) ? (a) : (b))
#define MAX(a, b) ((a) > (b) ? (a) : (b))

#define SAMPLES_PER_SECOND 48000

struct subchunk_header {
	char name[4];
	uint32_t size;
};

struct wav_info {
	char *filename;
	size_t filesize;

	char *file;
	uint32_t fmt_offset;
	uint32_t data_offset;
	uint16_t audio_format;
	uint16_t num_channels;
	uint32_t sample_rate;
	uint32_t byte_rate;
	uint16_t block_align;
	uint16_t bits_per_sample;

	int bytes_per_whole_sample;
	int num_samples;
	volatile int do_silent;
	volatile int song_offset;
	volatile float gain;
	volatile int do_echo;
	volatile int echo_offset;
};

int ctrl_fd;
int stream_fd;
struct libusb_device_handle *keyboard;
uint8_t endpoint_address;
struct wav_info song1, song2;
volatile int do_airhorn;

int16_t clamp(int32_t a)
{
	int32_t b = MIN(a, INT16_MAX);
	return MAX(b, INT16_MIN);
}

int32_t get_sample(struct wav_info wi, int ndx)
{
	float sample;
	static const float echo_weights[] = { 0.5, 0.25, 0.125 };

	if (wi.do_silent) {
		return 0;
	}

	sample = SAMPLE_AT_WO(wi, ndx);
	if (wi.do_echo) {
		for (int i = 0; i < sizeof(echo_weights)/sizeof(float); i++) {
			sample += echo_weights[i]*SAMPLE_AT_WO(wi, ndx + (i + 1)*wi.echo_offset);
		}
	}
	return (int32_t)(wi.gain*sample);
}

struct wav_info new_wav_info(const char *filename)
{
	uint32_t curr_offset = 12;  // skip first chunk
	struct wav_info wi = { 0 };
	uint32_t data_size;
	struct stat st;
	int fd;
	char *file;

	assert(lstat(filename, &st) >= 0);
	assert((fd = open(filename, O_RDONLY)) >= 0);
	assert((file = mmap(NULL, st.st_size, PROT_READ, MAP_PRIVATE, fd, 0)) != MAP_FAILED);
	assert(strncmp(file, "RIFF", 4) == 0 && strncmp(file + 8, "WAVE", 4) == 0);

	wi.filename = (char *)filename;
	wi.filesize = st.st_size;
	wi.file = file;

	while (!wi.fmt_offset || !wi.data_offset) {
		char *curr_subchunk = file + curr_offset;
		struct subchunk_header curr_header =
			*(struct subchunk_header *)curr_subchunk;

		curr_header.size = le32toh(curr_header.size);
		printf("%.*s: %d\n", 4, curr_header.name, curr_header.size);
		if (strncmp(curr_header.name, "fmt\040", 4) == 0) {
			wi.fmt_offset = curr_offset;
		} else if (strncmp(curr_header.name, "data", 4) == 0) {
			wi.data_offset = curr_offset;
		}

		curr_offset += 8 + curr_header.size;
	}

	assert(wi.fmt_offset && wi.data_offset);

	wi.audio_format = LE16TOH(FMT_OF(wi) + 8);
	wi.num_channels = LE16TOH(FMT_OF(wi) + 10);
	wi.sample_rate = LE32TOH(FMT_OF(wi) + 12);
	wi.byte_rate = LE32TOH(FMT_OF(wi) + 16);
	wi.block_align = LE16TOH(FMT_OF(wi) + 20);
	wi.bits_per_sample = LE16TOH(FMT_OF(wi) + 22);

	wi.bytes_per_whole_sample = wi.num_channels * (wi.bits_per_sample / 8);
	data_size = LE32TOH(DATA_OF(wi) + 4);
	wi.num_samples = data_size / wi.bytes_per_whole_sample;

	wi.do_silent = 0;
	wi.song_offset = 0;
	wi.gain = 1.0;
	wi.do_echo = 0;
	wi.echo_offset = 0;

	return wi;
}

void print_wav_info(struct wav_info wi)
{
	printf("audio_format=%d\n", wi.audio_format);
	printf("num_channels=%d\n", wi.num_channels);
	printf("sample_rate=%d\n", wi.sample_rate);
	printf("byte_rate=%d\n", wi.byte_rate);
	printf("block_align=%d\n", wi.block_align);
	printf("bits_per_sample=%d\n", wi.bits_per_sample);
	printf("bytes_per_whole_sample=%d\n", wi.bytes_per_whole_sample);
	printf("num_samples=%d\n", wi.num_samples);
}

void *keyboard_fn(void *arg)
{
	struct usb_keyboard_packet packet;
	int transferred;
	// char keystate[12];
	uint8_t prev_key0 = 0;
	uint8_t listen_for_song_offset = 0;
	int song_offset = 0;

	for (;;) {
		libusb_interrupt_transfer(keyboard, endpoint_address,
			(unsigned char *)&packet, sizeof(packet),
			&transferred, 0);

		// pprintf(keystate, "%02x %02x %02x", packet.modifiers, packet.keycode[0], packet.keycode[1]);
		// printf("%s\n", keystate);

		if (transferred != sizeof(packet)) {
			continue;
		}

		uint8_t key0 = packet.keycode[0];

		/*
		 * fn1: Play/pause song 1
		 * fn2: Play/pause song 2
		 * A/D: Scroll through song 1
		 * Left/right arrows: Scroll through song 2
		 */

		if (prev_key0 == key0) {
			continue;
		}
		prev_key0 = key0;

		switch (key0) {
		case KEY_F1:
			song1.do_silent = !song1.do_silent;
			printf("%s %s\n",
				song1.do_silent ? "Paused" : "Playing", song1.filename);
			break;
		case KEY_F2:
			song2.do_silent = !song2.do_silent;
			printf("%s %s\n",
				song2.do_silent ? "Paused" : "Playing", song2.filename);
			break;
		case KEY_F3:
			song1.do_silent = !song1.do_silent;
			song2.do_silent = !song2.do_silent;
			printf("%s %s. %s %s.\n",
				song1.do_silent ? "Paused" : "Playing", song1.filename,
				song2.do_silent ? "Paused" : "Playing", song2.filename);
			break;
		case KEY_A:
		case KEY_D:
		case KEY_LEFT:
		case KEY_RIGHT:
			listen_for_song_offset = key0;
			song_offset= 0;
			printf("Listening for song offset\n");
			break;
		case KEY_ENTER:
			switch (listen_for_song_offset) {
			case KEY_A:
				song1.song_offset -= song_offset * SAMPLES_PER_SECOND;
				printf("Rewinded %s by %ds\n", song1.filename, song_offset);
				break;
			case KEY_D:
				song1.song_offset += song_offset * SAMPLES_PER_SECOND;
				printf("Forwarded %s by %ds\n", song1.filename, song_offset);
				break;
			case KEY_LEFT:
				song2.song_offset -= song_offset * SAMPLES_PER_SECOND;
				printf("Rewinded %s by %ds\n", song2.filename, song_offset);
				break;
			case KEY_RIGHT:
				song2.song_offset += song_offset * SAMPLES_PER_SECOND;
				printf("Forwarded %s by %ds\n", song2.filename, song_offset);
				break;
			default:
				break;
			}
			listen_for_song_offset = song_offset = 0;
			break;
		case KEY_W:
			song1.gain *= GAIN_UP;
			printf("Set gain for %s to %.2f\n", song1.filename, song1.gain);
			break;
		case KEY_S:
			song1.gain *= GAIN_DOWN;
			printf("Set gain for %s to %.2f\n", song1.filename, song1.gain);
			break;
		case KEY_UP:
			song2.gain *= GAIN_UP;
			printf("Set gain for %s to %.2f\n", song2.filename, song2.gain);
			break;
		case KEY_DOWN:
			song2.gain *= GAIN_DOWN;
			printf("Set gain for %s to %.2f\n", song2.filename, song2.gain);
			break;
		case KEY_E:
			song1.do_echo = !song1.do_echo;
			song1.echo_offset = -(SAMPLES_PER_SECOND/8);
			printf("%s echo for %s\n",
				song1.do_echo ? "Starting" : "Stopped", song1.filename);
			break;
		case KEY_R:
			song2.do_echo = !song2.do_echo;
			song2.echo_offset = -(SAMPLES_PER_SECOND/8);
			printf("%s echo for %s\n",
				song2.do_echo ? "Starting" : "Stopped", song2.filename);
			break;
		case KEY_SPACE:
			do_airhorn = 1;
			printf("Playing airhorn\n");
			break;
		default:
			if (listen_for_song_offset && key0 >= KEY_1
				&& key0 <= KEY_0) {
				song_offset = song_offset*10 + (key0 == KEY_0 ? 0 : key0 - KEY_1 + 1);
			}
			break;
		}
	}
}

int main(int argc, char **argv)
{
	static const char ctrl_filename[] = "/dev/CTRL";
	static const char stream_filename[] = "/dev/STREAM";
	static const char airhorn_filename[] = "AIRHORN.wav";
	pthread_t keyboard_thread;
	struct wav_info airhorn;

	assert(argc == 3);
	char *song1_filename = argv[1];
	char *song2_filename = argv[2];

	printf("Mixer program started\n");

	if ((ctrl_fd = open(ctrl_filename, O_RDWR)) == -1) {
		fprintf(stderr, "could not open %s\n", ctrl_filename);
		return -1;
	}

	if ((stream_fd = open(stream_filename, O_RDWR)) == -1) {
		fprintf(stderr, "could not open %s\n", stream_filename);
		return -1;
	}

	if ((keyboard = openkeyboard(&endpoint_address)) == NULL) {
		fprintf(stderr, "Did not find keyboard\n");
		exit(1);
	}

	song1 = new_wav_info(song1_filename);
	song2 = new_wav_info(song2_filename);
	airhorn = new_wav_info(airhorn_filename);

	printf("\n");
	printf("SONG 1: %s\n", song1.filename);
	print_wav_info(song1);

	printf("\n");
	printf("SONG 2: %s\n", song2.filename);
	print_wav_info(song2);
	assert(song1.num_channels == 1 && song1.bits_per_sample == 16);
	assert(song2.num_channels == 1 && song2.bits_per_sample == 16);

	printf("\n");

	pthread_create(&keyboard_thread, NULL, keyboard_fn, NULL);

	int song1_ndx, song2_ndx, airhorn_ndx;
	airhorn_ndx = 0;
	for (song1_ndx = 0, song2_ndx = 0; ; song1_ndx += !song1.do_silent, song2_ndx += !song2.do_silent) {
		// printf("SAMPLE %d\n", i);
		int32_t sample1 = get_sample(song1, song1_ndx);
		int32_t sample2 = get_sample(song2, song2_ndx);

		// uint32_t sample = sample1 << 16 | sample2;
		// printf("sample=%x\t", sample);

		int32_t airhorn_sample = 0;
		if (do_airhorn) {
			airhorn_sample = SAMPLE_AT(airhorn, airhorn_ndx);
			airhorn_ndx++;
			if (airhorn_ndx >= airhorn.num_samples) {
				do_airhorn = airhorn_ndx = 0;
			}
		}

		stream_t stream = {
			.stream1 = clamp(sample1 + airhorn_sample*4),
			.stream2 = clamp(sample2)
		};
		stream_arg_t stream_arg = { .stream = stream };
		if (ioctl(stream_fd, STREAM_WRITE_STREAM, &stream_arg)) {
			perror("ioctl(STREAM_WRITE) failed");
			return -1;
		}
		// usleep(1);
		/*
		if (ioctl(stream_fd, STREAM_READ_STREAM, &stream_arg)) {
			perror("ioctl(STREAM_READ) failed");
			return -1;
		}*/
		// printf("stream1=%x\n", stream_arg.stream.stream1);
	}

	munmap(song1.file, song1.filesize);
	munmap(song2.file, song2.filesize);
	munmap(airhorn.file, airhorn.filesize);
	printf("Mixer program terminating\n");
	return 0;
}
