/* * Device driver for the VGA video generator
 *
 * A Platform device implemented using the misc subsystem
 *
 * Stephen A. Edwards
 * Columbia University
 *
 * References:
 * Linux source: Documentation/driver-model/platform.txt
 *               drivers/misc/arm-charlcd.c
 * http://www.linuxforu.com/tag/linux-device-drivers/
 * http://free-electrons.com/docs/
 *
 * "make" to build
 * insmod mixer.ko
 *
 * Check code style with
 * checkpatch.pl --file --no-tree vga_ball.c
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/platform_device.h>
#include <linux/miscdevice.h>
#include <linux/slab.h>
#include <linux/io.h>
#include <linux/of.h>
#include <linux/of_address.h>
#include <linux/fs.h>
#include <linux/uaccess.h>
#include "mixer.h"
#include <linux/wait.h>
#include <linux/interrupt.h>
#include <linux/of_irq.h>

#define CTRL_DRIVER_NAME "CTRL"
#define STREAM_DRIVER_NAME "STREAM"

/* Device registers */

//control slave
#define CTRL_L_GAIN(x) (x)
#define CTRL_R_GAIN(x) ((x)+1)

//stream slave
#define STREAM(x) (x)

//gain macros
#define L_GAIN 1
#define R_GAIN 1

/*
 * Information about our device
 */
struct ctrl_dev {
	struct resource res; /* Resource: our registers */
	void __iomem *virtbase; /* Where registers can be accessed in memory */
    ctrl_gain_t gain;
} ctrl_dev;

struct stream_dev {
	struct resource res; /* Resource: our registers */
	void __iomem *virtbase; /* Where registers can be accessed in memory */
	stream_t stream;
	int irq_num;
	int ready;
} stream_dev;

DECLARE_WAIT_QUEUE_HEAD(wq);

irqreturn_t irq_handler(int irq, void *dev_id)
{
	if (irq != stream_dev.irq_num) {
		return IRQ_NONE;
	}

	stream_dev.ready = (ioread32(stream_dev.virtbase) == 0xe);
	wake_up_interruptible(&wq);
	return IRQ_HANDLED;
}

static void write_gain(ctrl_gain_t *gain) 
{
	iowrite8(gain->l_gain, CTRL_L_GAIN(ctrl_dev.virtbase) );
	iowrite8(gain->r_gain, CTRL_R_GAIN(ctrl_dev.virtbase) );
	ctrl_dev.gain = *gain;
}

static void write_stream(stream_t *stream)
{
	wait_event_interruptible(wq, stream_dev.ready);
	// wait_event_interruptible_timeout(wq, stream_dev.ready, 10);
	iowrite32(stream->stream1 << 16 | stream->stream2, stream_dev.virtbase);
	// iowrite16(stream->stream2, stream_dev.virtbase + 2);
	stream_dev.stream = *stream;
}

static void read_stream(stream_t *stream)
{
	stream->stream1 = ioread16(stream_dev.virtbase);
	// stream->stream2 = ioread16(stream_dev.virtbase + 2);
}

/*
 * Handle ioctl() calls from userspace
 */
static long ctrl_ioctl(struct file *f, unsigned int cmd, unsigned long arg)
{
	ctrl_arg_t ctrl;

	switch (cmd) {
	case CTRL_WRITE_GAIN:
		if (copy_from_user(&ctrl, (ctrl_arg_t *) arg,
				   sizeof(ctrl_arg_t)))
			return -EACCES;
		write_gain(&ctrl.gain);
		break;
	default:
		return -EINVAL;
	}

	return 0;
}

static long stream_ioctl(struct file *f, unsigned int cmd, unsigned long arg)
{
	stream_arg_t stream;

	switch (cmd) {
	case STREAM_WRITE_STREAM:
		if (copy_from_user(&stream, (stream_arg_t *) arg,
				   sizeof(stream_arg_t)))
			return -EACCES;
		write_stream(&stream.stream);
		break;
	case STREAM_READ_STREAM:
		read_stream(&stream.stream);
		if (copy_to_user((stream_arg_t *)arg, &stream,
			sizeof(stream_arg_t))) {
			return -EACCES;
		}
		break;
	default:
		return -EINVAL;
	}

	return 0;
}

/* The operations our device knows how to do */
static const struct file_operations ctrl_fops = {
	.owner		= THIS_MODULE,
	.unlocked_ioctl = ctrl_ioctl,
};

static const struct file_operations stream_fops = {
	.owner		= THIS_MODULE,
	.unlocked_ioctl = stream_ioctl,
};

/* Information about our device for the "misc" framework -- like a char dev */
static struct miscdevice ctrl_misc_device = {
	.minor		= MISC_DYNAMIC_MINOR,
	.name		= CTRL_DRIVER_NAME,
	.fops		= &ctrl_fops,
};

static struct miscdevice stream_misc_device = {
	.minor		= MISC_DYNAMIC_MINOR,
	.name		= STREAM_DRIVER_NAME,
	.fops		= &stream_fops,
};

/*
 * Initialization code: get resources (registers)
 */
static int __init ctrl_probe(struct platform_device *pdev)
{
	int ret;

	/* Register ourselves as a misc device: creates /dev/vga_ball */
	ret = misc_register(&ctrl_misc_device);

	/* Get the address of our registers from the device tree */
	ret = of_address_to_resource(pdev->dev.of_node, 0, &ctrl_dev.res);
	if (ret) {
		ret = -ENOENT;
		goto out_deregister;
	}

	/* Make sure we can use these registers */
	if (request_mem_region(ctrl_dev.res.start, resource_size(&ctrl_dev.res),
		CTRL_DRIVER_NAME) == NULL) {
		ret = -EBUSY;
		goto out_deregister;
	}

	/* Arrange access to our registers */
	ctrl_dev.virtbase = of_iomap(pdev->dev.of_node, 0);
	if (ctrl_dev.virtbase == NULL) {
		ret = -ENOMEM;
		goto out_release_mem_region;
	}

	return 0;

out_release_mem_region:
	release_mem_region(ctrl_dev.res.start, resource_size(&ctrl_dev.res));
out_deregister:
	misc_deregister(&ctrl_misc_device);
	return ret;
}

/* Clean-up code: release resources */
static int ctrl_remove(struct platform_device *pdev)
{
	iounmap(ctrl_dev.virtbase);
	release_mem_region(ctrl_dev.res.start, resource_size(&ctrl_dev.res));
	misc_deregister(&ctrl_misc_device);
	return 0;
}

static int __init stream_probe(struct platform_device *pdev)
{
	int ret, irq;

	/* Register ourselves as a misc device: creates /dev/vga_ball */
	ret = misc_register(&stream_misc_device);

	/* Get the address of our registers from the device tree */
	ret = of_address_to_resource(pdev->dev.of_node, 0, &stream_dev.res);
	if (ret) {
		ret = -ENOENT;
		goto out_deregister;
	}

	/* Make sure we can use these registers */
	if (request_mem_region(stream_dev.res.start, resource_size(&stream_dev.res),
		STREAM_DRIVER_NAME) == NULL) {
		ret = -EBUSY;
		goto out_deregister;
	}

	/* Arrange access to our registers */
	stream_dev.virtbase = of_iomap(pdev->dev.of_node, 0);
	if (stream_dev.virtbase == NULL) {
		ret = -ENOMEM;
		goto out_release_mem_region;
	}

	irq = irq_of_parse_and_map(pdev->dev.of_node, 0);
	stream_dev.irq_num = irq;

	pr_alert("irq_of_parse_and_map worked\n");

	ret = request_irq(irq, irq_handler, 0, "csee4840_vga_ball", NULL);
	pr_alert("request_irq finished\n");
	if (ret) {
		pr_err("request_irq failed: %d\n", ret);
		ret = -ENOENT;
		goto out_release_mem_region;
	}

	stream_dev.ready = 1;

	return 0;

out_release_mem_region:
	release_mem_region(stream_dev.res.start, resource_size(&stream_dev.res));
out_deregister:
	misc_deregister(&stream_misc_device);
	return ret;
}

/* Clean-up code: release resources */
static int stream_remove(struct platform_device *pdev)
{
	free_irq(stream_dev.irq_num, NULL);
	iounmap(stream_dev.virtbase);
	release_mem_region(stream_dev.res.start, resource_size(&stream_dev.res));
	misc_deregister(&stream_misc_device);
	return 0;
}

/* Which "compatible" string(s) to search for in the Device Tree */
#ifdef CONFIG_OF
static const struct of_device_id ctrl_of_match[] = {
	{ .compatible = "csee4840,ctrl-1.0" },
	{},
};
MODULE_DEVICE_TABLE(of, ctrl_of_match);
static const struct of_device_id stream_of_match[] = {
	{ .compatible = "csee4840,vga_ball-1.0" },
	{},
};
MODULE_DEVICE_TABLE(of, stream_of_match);
#endif

/* Information for registering ourselves as a "platform" driver */
static struct platform_driver ctrl_driver = {
	.driver	= {
		.name	= CTRL_DRIVER_NAME,
		.owner	= THIS_MODULE,
		.of_match_table = of_match_ptr(ctrl_of_match),
	},
	.remove	= __exit_p(ctrl_remove),
};

static struct platform_driver stream_driver = {
	.driver	= {
		.name	= STREAM_DRIVER_NAME,
		.owner	= THIS_MODULE,
		.of_match_table = of_match_ptr(stream_of_match),
	},
	.remove	= __exit_p(stream_remove),
};

/* Called when the module is loaded: set things up */
static int __init mixer_init(void)
{
	int ret;

	pr_info(CTRL_DRIVER_NAME ": init\n");
	ret = platform_driver_probe(&ctrl_driver, ctrl_probe);
	if (ret < 0)
		return ret;
	pr_info(STREAM_DRIVER_NAME ": init\n");
	return platform_driver_probe(&stream_driver, stream_probe);
}

/* Calball when the module is unloaded: release resources */
static void __exit mixer_exit(void)
{
	platform_driver_unregister(&ctrl_driver);
	platform_driver_unregister(&stream_driver);
	pr_info(CTRL_DRIVER_NAME ": exit\n");
	pr_info(STREAM_DRIVER_NAME ": exit\n");
}

module_init(mixer_init);
module_exit(mixer_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Joy, Oli, Harrison, Joshua. Columbia University");
MODULE_DESCRIPTION("FPGA DJ");
